\newpage

# Virtualisation de stockage avec LVM

## Exercice 1

### Quel est l'espace disponible sur la partition où est monté le système de fichiers racine (/) ?

En totalité, 18GB sont disponibles dont 14GB sont déjà occupés.

```bash
/dev/mapper/ubuntu--vg-ubuntu--lv   18G   14G  3.4G  81% /
```

### Investigation avec `lsblk`

```bash
NAME                      MAJ:MIN RM  SIZE RO TYPE MOUNTPOINTS
loop0                       7:0    0   62M  1 loop /snap/core20/1587
loop1                       7:1    0 63.9M  1 loop /snap/core20/2182
loop2                       7:2    0   87M  1 loop /snap/lxd/27428
loop3                       7:3    0   87M  1 loop /snap/lxd/27948
loop4                       7:4    0 40.4M  1 loop /snap/snapd/20671
loop5                       7:5    0 39.1M  1 loop /snap/snapd/21184
sda                         8:0    0   20G  0 disk 
├─sda1                      8:1    0    1M  0 part 
├─sda2                      8:2    0  1.8G  0 part /boot
└─sda3                      8:3    0 18.2G  0 part 
  └─ubuntu--vg-ubuntu--lv 253:0    0 18.2G  0 lvm  /
sdb                         8:16   0   20G  0 disk 
sdc                         8:32   0   20G  0 disk 
sr0                        11:0    1  1.4G  0 rom
```

### Inspection des partitions avec `fdisk -l /dev/<partition>`


#### Partition `/dev/sda1` 

```bash
Disk /dev/sda1: 1 MiB, 1048576 bytes, 2048 sectors
Units: sectors of 1 * 512 = 512 bytes
Sector size (logical/physical): 512 bytes / 512 bytes
I/O size (minimum/optimal): 512 bytes / 512 bytes
```

#### Partition `/dev/sda2` 

```bash
Disk /dev/sda2: 1.77 GiB, 1902116864 bytes, 3715072 sectors
Units: sectors of 1 * 512 = 512 bytes
Sector size (logical/physical): 512 bytes / 512 bytes
I/O size (minimum/optimal): 512 bytes / 512 bytes
```

#### Partition `/dev/sda3` 

```bash
Disk /dev/sda3: 18.23 GiB, 19569573888 bytes, 38221824 sectors
Units: sectors of 1 * 512 = 512 bytes
Sector size (logical/physical): 512 bytes / 512 bytes
I/O size (minimum/optimal): 512 bytes / 512 bytes
```

### Création de la structure de stockage avec LVM

#### Création d'un volume groupe constitué du second disque

```bash
vgcreate vgb /dev/sdb
```

#### Création d'un volume logique de la taille du volume groupe

```bash
lvcreate -n vol -l 100%VG vgb
```

#### Formattez le volume logique en `ext4` 

```bash
mkfs.ext4 /dev/vgb/vol
```

#### Montez le volume logique dans le répertoire de votre choix

```bash
sudo mount /dev/vgb/vol fs/
```

#### Quelle est la taille d'un _physical extent_ ?

La taille d'un _physical extent_ est de 4096KB ou 4 MiB.

```bash
sudo vgdisplay
```

```bash
--- Volume group ---                                                                                                                                        
VG Name               vgb                                                                                                                                   
System ID                                                                                                                                                   
Format                lvm2                                                                                                                                  
Metadata Areas        1                                                                                                                                     
Metadata Sequence No  2                                                                                                                                     
VG Access             read/write                                                                                                                            
VG Status             resizable                                                                                                                             
MAX LV                0                                                                                                                                     
Cur LV                1                                                                                                                                     
Open LV               1                                                                                                                                     
Max PV                0                                                                                                                                     
Cur PV                1                                                                                                                                     
Act PV                1                                                                                                                                     
VG Size               <20.00 GiB                                                                                                                            
PE Size               4.00 MiB                                                                                                                              
Total PE              5119                                                                                                                                  
Alloc PE / Size       5119 / <20.00 GiB                                                                                                                     
Free  PE / Size       0 / 0                                                                                                                                 
VG UUID               H77qrp-fwPT-WtBS-3Ixf-DliK-fxnC-0EJWz9                                                                                                
```

#### Quelle est la capacité du volume groupe créé précédemment ?

La capacité du volume logique est de 20 GiB.

```bash
sudo lvdisplay
```

```bash
--- Logical volume ---
LV Path                /dev/vgb/vol
LV Name                vol
VG Name                vgb
LV UUID                2g80t7-uZBx-UgMz-p8yY-WNSD-80wd-Y54VaF
LV Write Access        read/write
LV Creation host, time ISC-virtu-45, 2024-04-03 20:02:53 +0000
LV Status              available
# open                 1
LV Size                <20.00 GiB
Current LE             5119
Segments               1
Allocation             inherit
Read ahead sectors     auto
- currently set to     256
Block device           253:1
```

#### Peut-on agrandir le volume groupe en lui ajoutant le disque sur lequel se trouve l'OS ?

En l'occurrence cela n'est sûrement pas possible car l'OS se trouve dans un
autre volume groupe.

#### Aurait-il été possible d'étendre le volume logique `ubuntu-lv``avec le contenu des disques `sdb` et `sdc` ?

Possiblement avec le contenu de `sdc` mais pas celui de `sdb` car il se trouve
dans un autre volume groupe.

#### Étendez le volume groupe créé précédemment avec le disque restant

```bash
sudo vgextend vgb /dev/sdc
```

#### Étendez l'espace du volume logique afin qu'il utilise l'entièreté du volume groupe

```bash
sudo lvextend -l 100%VG /dev/vgb/vol
```

#### Étendez finalement le système de fichiers contenu dans ce volume logique

```bash
sudo resize2fs /dev/vgb/vol
```

## Exercice 2

### Pourquoi doit-on définir une taille pour un volume de type snapshot ?

Il est nécessaire de définir une taille pour un volume de type snapshot car
ce volume est censé pouvoir stocker **tous** les blocs originaux des fichiers
qui seront plus tard modifiés dans le volume original.

### Création d'un snapshot

Une erreur survient lorsqu'on essaye d'exécuter la commande ci-dessous car notre
volume groupe n'a plus d'espace disponible.

```bash
sudo lvcreate -s -n snap -L 10G /dev/vgb/vol
```

Par conséquent, nous devons réduire la taille du volume groupe à 50% de la taille
actuelle.

```bash
student@ISC-virtu-45:~$ sudo lvreduce -l 50%VG /dev/vgb/vol
WARNING: Reducing active and open logical volume to <20.00 GiB.
THIS MAY DESTROY YOUR DATA (filesystem etc.)
```

Il s'avère qu'il n'est pas possible de modifier à chaud la capacité d'un volume
logique contenant un système de fichiers `ext4`.

Par conséquent il est nécessaire d'effectuer les étapes dans cet ordre afin de
redimensionner le volume logique :

1. Démonter le système de fichiers
  ```bash
  sudo umount fs/
  ```
2. Redimensionner la taille du système de fichiers
  - Avant de pouvoir rédimensionner le système de fichiers, j'ai dû exécuté
  la commande `e2fsck -f /dev/vgb/vol` afin de vérifier l'état du système de
  fichiers
  ```bash
  sudo resize2fs /dev/vgb/vol 20G
  ```
3. Réduire la taille du volume logique
  ```bash
  student@ISC-virtu-45:~$ sudo lvreduce -l 50%VG /dev/vgb/vol
  WARNING: Reducing active logical volume to <20.00 GiB.
  THIS MAY DESTROY YOUR DATA (filesystem etc.)
  Do you really want to reduce vgb/vol? [y/n]: y
  Size of logical volume vgb/vol changed from 39.99 GiB (10238 extents) to <20.00 GiB (5119 extents).
  Logical volume vgb/vol successfully resized.
  ```
4. Vérifier que le système de fichiers est valide (commande `fsck.ext4`)
  ```bash
  student@ISC-virtu-45:~$   sudo fsck.ext4 /dev/vgb/vol
  e2fsck 1.46.5 (30-Dec-2021)
  /dev/vgb/vol: clean, 11/1310720 files, 126322/5241856 blocks
  ```
5. Monter le système de fichiers
  ```bash
  sudo mount /dev/vgb/vol fs/
  ```

Finalement la création du snapshot `snap` a fonctionné :

```bash
student@ISC-virtu-45:~$ sudo lvcreate -s -n snap -L 10G /dev/vgb/vol
Logical volume "snap" created.
```

### Montez le snapshot

```bash
sudo mount /dev/vgb/snap snapshots/
```

#### Quel est son contenu par rapport au volume "original" ?

Il semblerait que le contenu entre les deux volumes soit le même.

```bash
student@ISC-virtu-45:~/fs$ ls -altr
total 24
drwx------ 2 root    root    16384 Apr  4 10:46 lost+found
drwxr-xr-x 3 root    root     4096 Apr  4 10:46 .
drwxr-x--- 8 student student  4096 Apr  4 10:58 ..
```

```bash
student@ISC-virtu-45:~/snapshots$ ls -altr
total 24
drwx------ 2 root    root    16384 Apr  4 10:46 lost+found
drwxr-xr-x 3 root    root     4096 Apr  4 10:46 .
drwxr-x--- 8 student student  4096 Apr  4 10:58 ..
```

### Populez le volume original

Nous allons maintenant créer des nouveaux fichiers dans le volume "original"
afin de voir les divergences avec le volume `snap`. Un fichier de 10% de la
taille du snapshot sera aussi créé (1GB).

```bash
student@ISC-virtu-45:~/fs$ sudo dd if=/dev/urandom of=bigfile bs=1M count=1K
1024+0 records in
1024+0 records out
1073741824 bytes (1.1 GB, 1.0 GiB) copied, 3.12577 s, 344 MB/s
```

Contenu du volume :

```bash
student@ISC-virtu-45:~/fs$ ls -altr
total 1048608
drwx------ 2 root    root         16384 Apr  4 10:46 lost+found
drwxr-x--- 8 student student       4096 Apr  4 11:07 ..
-rw-r--r-- 1 root    root            10 Apr  4 11:07 toto
drwxr-xr-x 3 root    root          4096 Apr  4 11:07 .
-rw-r--r-- 1 root    root    1073741824 Apr  4 11:07 bigfile
```

### Inspectez l'espace utilisé par le snapshot

Il semblerait qu'à présent 10% du volume `snap` soit utilisé ce qui serait
en correlation avec le fichier `bigfile` créé (1GB).

```bash
student@ISC-virtu-45:~/snapshots$ sudo lvs
LV        VG        Attr       LSize   Pool Origin Data%  Meta%  Move Log Cpy%Sync Convert
ubuntu-lv ubuntu-vg -wi-ao----  18.22g                                                    
snap      vgb       swi-aos---  10.00g      vol    10.04                                  
vol       vgb       owi-aos--- <20.00g
```

### Comparaison entre `/dev/vgb/vol` et `/dev/vgb/snap` 

En comparant le contenu de `fs/` et `snapshots/`, il semblerait que les fichiers
créer dans le volume `/dev/vgb/vol` n'apparaissent pas du côté du volume
`/dev/vgb/snap` monté au niveau de `~/snapshots/`.

#### `/dev/vgb/vol`

```bash
student@ISC-virtu-45:~/fs$ ls -altr
total 1048608
drwx------ 2 root    root         16384 Apr  4 10:46 lost+found
drwxr-x--- 8 student student       4096 Apr  4 11:07 ..
-rw-r--r-- 1 root    root            10 Apr  4 11:07 toto
drwxr-xr-x 3 root    root          4096 Apr  4 11:07 .
-rw-r--r-- 1 root    root    1073741824 Apr  4 11:07 bigfile
```

#### `/dev/vgb/snap`

```bash
student@ISC-virtu-45:~/snapshots$ ls -altr
total 24
drwx------ 2 root    root    16384 Apr  4 10:46 lost+found
drwxr-xr-x 3 root    root     4096 Apr  4 10:46 .
drwxr-x--- 8 student student  4096 Apr  4 11:07 ..
```

### Restauration du volume `/dev/vgb/vol` à l'état du snapshot

#### Peut-on réaliser cette opération à chaud (_online_) ?

Non, cela n'est pas possible. Il nécessaire que les volumes `/dev/vgb/vol` et
`/dev/vgb/snap` soient les deux démontés.

```bash
sudo umount fs/
```

```bash
sudo umount snapshots/
```

#### _Merge_ du snapshot `/dev/vgb/snap` dans `/dev/vgb/vol`

```bash
sudo lvconvert --merge vgb/snap
```

#### Quelle a été la durée approximative de cette opération ?

Le _merge_ a duré approximativement 40 secondes.

#### Comment pouvez-vous valider que tout s'est bien passé ?

En remontant le volume `/dev/vgb/vol`, nous pouvons inspectez les fichiers
présents :

```bash
sudo mount /dev/vgb/vol fs/
```

```bash
student@ISC-virtu-45:~/fs$ ls -altr
total 24
drwx------ 2 root    root    16384 Apr  4 10:46 lost+found
drwxr-xr-x 3 root    root     4096 Apr  4 10:46 .
drwxr-x--- 8 student student  4096 Apr  4 11:07 ..
```

Nous pouvons par conséquent remarqué que l'état du volume `/dev/vgb/vol` a bien
été restauré à l'état du snapshot.

#### Quels sont les avantages principaux de LVM par dessus des disques physiques, plutôt que d'y installe directement un FS comme `ext4` ? 

En ce qui concerne notamment le système de fichiers `ext4` qui ne supporte pas
nativement des snapshots, LVM permet de palier à ce problème en rajoutant
cette fonctionnalité à `ext4`. LVM permet aussi de créer plusieurs des groupes
de volumes qui peuvent "encapsuler" plusieurs volumes physiques et le présenter
comme un seul et unique volume.

#### Pourquoi est-ce que la commande `lvconvert --merge` prend un certain temps à se terminer ?

Plus il y a de différences entre le volume original et le snapshot (c'est-à-dire
le moment où le volume snapshot a été créé), plus cela prendra de temps pour
rétablir les différences entre le volume original et le snapshot.

## Exercice 3

### Aggrégation de tous les disques en seul volume groupe

#### Démontez les volumes

- `/dev/vgb/vol` 
  ```bash
  sudo umount fs/
  ```
- `/dev/vgb/snap`
  ```bash
  sudo umount snapshots/
  ```

#### Suppression du volume logique `/dev/vgb/vol`

```bash
student@ISC-virtu-45:~$ sudo lvremove /dev/vgb/vol
Do you really want to remove and DISCARD active logical volume vgb/vol? [y/n]: y
Logical volume "vol" successfully removed
```

#### Suppression du volume groupe `/dev/vgb` 

```bash
student@ISC-virtu-45:~$ sudo vgremove vgb
Volume group "vgb" successfully removed
```

#### Rajout des disques physiques `/dev/sdb` et `/dev/sdc` dans le volume groupe `ubuntu-vg`

```bash
student@ISC-virtu-45:~$ sudo vgextend ubuntu-vg /dev/sdb /dev/sdc
Volume group "ubuntu-vg" successfully extended
```

#### Extension de la taille du volume logique `ubuntu-lv` jusqu'à la taille du volume groupe `ubuntu-vg`

```bash
student@ISC-virtu-45:~$ sudo lvextend -l 100%VG /dev/ubuntu-vg/ubuntu-lv
Size of logical volume ubuntu-vg/ubuntu-lv changed from 18.22 GiB (4665 extents) to 58.21 GiB (14903 extents).
Logical volume ubuntu-vg/ubuntu-lv successfully resized.
```

#### Extension de la taille du FS contenu dans le volume logique `/dev/ubuntu-vg/ubuntu-lv` 

```bash
student@ISC-virtu-45:~$ sudo resize2fs /dev/ubuntu-vg/ubuntu-lv 
resize2fs 1.46.5 (30-Dec-2021)
Filesystem at /dev/ubuntu-vg/ubuntu-lv is mounted on /; on-line resizing required
old_desc_blocks = 3, new_desc_blocks = 8
The filesystem on /dev/ubuntu-vg/ubuntu-lv is now 15260672 (4k) blocks long.
```

#### Résultat des courses

En exécutant la commande `df -h`, nous pouvons bien remarqué que les disques
`sdb` et `sdc` on été fusionné dans un seul et unique groupe, volume et finalement
FS ce qui nous donne un FS de la capacité du volume et groupe de ~ 60GB

```bash
student@ISC-virtu-45:~$ df -h
Filesystem                         Size  Used Avail Use% Mounted on
tmpfs                              794M  1.3M  793M   1% /run
/dev/mapper/ubuntu--vg-ubuntu--lv   58G   14G   42G  25% /
tmpfs                              3.9G     0  3.9G   0% /dev/shm
tmpfs                              5.0M     0  5.0M   0% /run/lock
/dev/sda2                          1.8G  252M  1.4G  16% /boot
tmpfs                              3.9G     0  3.9G   0% /run/qemu
tmpfs                              794M  4.0K  794M   1% /run/user/1001
```
