\newpage

# Docker Basics

## Exercice 1

Localisez le container correspondant au 1er exercice à l'aide de la commande
`docker search` :

```bash
docker search hepia
```

En l'occurrence, l'exercice 1 correspond au container `hepia/docker_ex01` 

```bash
NAME                 DESCRIPTION                                     STARS     OFFICIAL
hepia/docker_ex01                                                    0         
hepia/docker_ex03                                                    0         
hepia/hepia-pandoc   convert markdown to pdf with custom template…   0         
hepia/docker_ex02                                                    0         
hepia/docker_ex04                                                    0         
hepia/ubuntu-java    Ubuntu 19.04 with openjdk and other tools ar…   0         
hepia/docker_ex05                                                    0         
hepia/hbvar-openms   Open-source software for mass spectrometry a…   0         
peiryd/hepia                                                         0
```

Sans exécuter le shell intéractif dans le container, effectuez les actions
suivantes :

1. Listez les fichiers se trouvant à la racine
    ```bash
    docker run --workdir / --rm hepia/docker_ex01 ls -ltr
    ```

    - Filtrez pour afficher le seul "vrai" fichier (et pas caché)
        ```bash
        docker run --workdir / --rm hepia/docker_ex01 find . -maxdepth 1 -type f -not -path '*/.*'
        ```

2. Affichez le contenu du fichier
    ```bash
    docker run --workdir / --rm hepia/docker_ex01 cat flag
    ```

- Suite aux deux opérations ci-dessus, combien de containers ont été créés et
quels sont leurs noms et IDs ?
    - Pour chacune des commandes précédentes un nouveau container a été lancé
    à chaque fois (2 en tout). Vu que j'ai explicitement spécifié l'argument
    `--rm` lors de l'exécution de `docker run`, les containers créés ont été
    supprimés à la fin de leur exécution. Si cela n'avait pas été fait, la
    commande `docker ps` permet de lister tous les containers en cours d'exécution
    ainsi que leur IDs, états et aux informations utiles.

- Quel est l'état des containers ?
    - Les containers ont été supprimés.

- Quel est l'ID de l'image utilisée par les containers que vous venez d'exécuter ?
    - Afin de découvrir cette information, nous devons exécuter la commande
    `docker images` qui permettra de lister les images disponibles et quelques
    informations associées. En l'occurrence, l'ID de l'image `hepia/docker_ex01`
    est `cae01e578e53`.

- Est-ce que l'ID de l'image sera identique chez vos collègues de classe ?
    - Oui, car l'ID de l'image correspond aux 12 premiers octets du _digest_
    SHA-256 de l'image. Cela peut se vérifier avec la commande ci-dessous :
        ```bash
        docker inspect hepia/docker_ex01 | grep "cae01e578e53"
        ```

- Est-ce que les noms et IDs des containers seront identiques chez vos collègues ?
    - Non, ces identifiants sont aléatoires

- Sur quelle machine se trouvent les images `docker` ?
    - Les images se trouvent sur le serveur exécutant le daemon `docker`, en
    l'occurrence c'est la VM de l'école.

## Exercice 2

Exécutez le container `hepia/docker_ex03` avec un shell (`sh`) intéractif et
créez ensuite le fichier `fantasio` dans le répertoire `/ex03/` avec la commande
`touch`.

```bash
docker run -it --rm hepia/docker_ex03 /bin/sh
```

```bash
cd ex03
```

```bash
touch fantasio
```

- Que se passe-t-il quand vous quittez le shell avec `Ctrl-d` ?
    - Le container est arrêté mais n'est pas supprimé (en l'occurrence il le
    sera je l'exécute avec l'option `--rm`).

Exécutez un nouveau container basé sur la même image que précédemment et nommez-la
`zorglub`.

```bash
docker run -it --name zorglub hepia/docker_ex03 /bin/sh
```

- Est-ce que le fichier `/ex03/fantasio` existe ou pas ? Pourquoi ?
    - Non, le fichier en question n'existera pas dans ce nouveau container
    car le container `zorglub` n'est qu'une instance de l'image `hepia/docker_ex03`.
    Ce container n'a rien en commun avec le container précédent à part l'image
    **_parent_** qu'ils ont en commun.

Dans le container `zorglub`, créez le fichier `/ex03/fantasio`, puis pressez la
combinaison de touches `Ctrl-p` `Ctrl-q`.

- Quel est l'état (stoppé, en exécution, etc...) du container `zorglub` ?
    - À l'aide de cette combinaison de touches, nous avons simplement détaché
    le terminal intéractif du container. Par conséquent, le container est toujours
    en cours d'exécution.

    - Le rôle de cette combinaison est de se détacher du shell intéractif du
    container sans pour autant terminé son exécution.

Attachez-vous (`docker attach`) alors au container `zorglub` ?

```bash
docker attach zorglub
```

- Est-ce que le fichier `/ex03/fantasio` existe encore ?
    - Oui, le fichier est présent.

Dans un nouveau terminal, attachez-vous à nouveau au container `zorglub`. Placez
à côte à côte les deux terminaux dans lesquels vous êtes attaché au container
`zorglub` et tapez des caractères au clavier.

- Qu'observez-vous et que pouvez-vous donc en déduire ?
    - Nous pouvons donc en déduire que nous pouvons attacher un nombre arbitraire
    de terminaux à un container cependant celui-ci ne sera dupliqué donc cela
    revient à avoir une multitude de vues d'un seul _processus_.

![2 terminals, 1 container (iykyk)](../figs/docker_two_terminals.png){width=60%}

Exécutez la commande `docker exec -it zorglub sh`

- Est-ce que le fichier `/ex03/fantasio` existe toujours ? Combien de shells
sont en cours d'exécution (aide : `ps`) ?
    - Oui, le fichier existe toujours. En exécutant la commande `ps -ef`, on
    s'aperçoit que deux shells sont en cours d'exécution (`/bin/sh` et `sh`).

- Quelle est la différence entre les commandes `attach` et `exec` ?
    - La différence entre ces deux commandes sont les suivantes :
        - `attach` permet de se "réattacher" au processus qui a été lancé
        au démarrage du container
        - `exec` permet de lancer un nouveau processus dans un container démarré
        (ex.: lancer un nouveau shell afin d'inspécter son système de fichiers).

Terminez le shell (p.ex. avec `Ctrl-d`) dans lequel vous aviez exécuté `docker exec ...`
auparavant.

- Est-ce que le container `zorglub` est toujours en cours d'exécution ?
    - Oui, il l'est toujours.

Exécutez un nouveau shell avec la commande `docker exec -it zorglub sh`,
puis allez dans le terminal où vous aviez réalisé le `docker attach` plus haut.
Terminez alors ce shell avec `Ctrl-d`.

- Est-ce que le container `zorglub` est toujours en cours d'exécution ? Que
pouvez-vous en conclure ?
    - En ayant fait cette manipulation, on s'aperçoit que nous avons non seulement
    arrêté le processus lancé par `docker exec ...`, nous avons aussi complètement
    arrêté le container ainsi que tous les vues que nous possédions vers celui-ci.
    La conclusion est (comme je l'ai mentionné précédemment) que la commande
    `docker attach` permet de nous "attacher" au processus original du container
    et par conséquent, si on le quitte, cela va arrêter le container.

Démarrez le container `zorglub` avec la commande `start -ia` (_attach_ + _interactive_).

- Le fichier `/ex03/fantasio` existe-t-il toujours ?
    - Oui, le fichier existe toujours.

## Exercice 3

Exécutez un container basé sur l'image Alpine 3.17 en spécifiant de le détruire
une fois terminé.

```bash
docker run -it --rm alpine:3.17
```

Vérifiez que vous utilisez la bonne distribution en inspectant le contenu du
fichier `/etc/os-release`.

```bash
cat /etc/os-release
```

```bash
NAME="Alpine Linux"
ID=alpine
VERSION_ID=3.17.7
PRETTY_NAME="Alpine Linux v3.17"
HOME_URL="https://alpinelinux.org/"
BUG_REPORT_URL="https://gitlab.alpinelinux.org/alpine/aports/-/issues"
```

Dans ce container, exécutez la commande `top`. Inspéctez ensuite le processus
`top` à l'intérieur du container et sur la machine hôte (via `ps aux`).

- Selon votre observation, est-ce qu'un processus appartenant à `root` dans le
container appartient également à root en dehors du container ?
    - Cela dépend si le daemon `dockerd` s'exécute en tant que `root` ou non.
    S'il l'est, alors le processus `root` dans le container le sera aussi en
    dehors. (_weird ngl, gotta verify that_).
    - Non, ce processus n'appartient pas à `root` en dehors du container.

Pour rappel, `root` possède le **UID** 0.

- Pourquoi est-il dangereux qu'un processus UID 0 dans le container soit également
UID 0 sur la machine hôte ?
    - Cela impliquerait que ce processus aurait tous les droits sur la machine hôte.

- Quelle différence remarquez-vous lorsque vous exécutez la commande `ps aux`
dans le container et sur la machine hôte ?
    - En ce qui concerne le container, on ne voit que les processus `/bin/sh`
    et `ps aux` s'exécuter à l'inverse de la machine hôte où un nombre bien plus
    conséquent de processus s'exécutent.

- À votre avis, quelle est la raison de cette différence ?
    - J'imagine que le processus du container est complètement isolé de son
    environnement "parent" pour des mesures de sécurité d'où le fait qu'on n'ait
    pas accès aux processus s'exécutant sur la machine hôte.

À l'intérieur du container, terminez `top`, puis installez et exécutez le programme
`asciiquarium`. Pour info, Alpine utilise le _package manager_ `apk`. Les
équivalents Alpine de `apt-get update` et `apt-get install` sont `apk update`
et `apk add`. Vous pouvez lister tous les paquets disponibles avec `apk list`.

- Quel est le PID de ce processus dans le container ? Quel est le PID de ce même
processus sur la machine hôte et pourquoi est-il différent à votre avis ?
    - Dans le container, le PID du processus est 22. Du côté de la mâchine hôte,
    le PID est 16125. Je suppose que la raison pour laquelle ces deux valeurs
    sont différentes est que le container exécute un OS à part entière d'où le
    fait que de son point de vue, il est le seul _responsable_ de ce processus.
    (_prolly bs_).

Vous allez maintenant investiguer empiriquement le comportement des containers.

- Quel est le processus portant le PID 1 dans le container ? Quel est le processus
portant le même numéro de PID sur la machine hôte ?
    - Dans le container, c'est le processus `/bin/sh` qui porte le PID 1 (cela
    est facilement visualisable grâce à `top`). Dans le cas de la machine hôte,
    en exécutant la commande `pstree -p | head`, on s'aperçoit que le PID 1
    correspond au processus `systemd`.

- Pour généraliser, quel est le processus portant le PID 1 dans le container
(c-à-d, quel que soit le container exécuté) ? Vous pouvez déduire empiriquement
en exécutant d'autres containers avec des programmes différents.
    - En ayant essayé avec l'image `isrkhn/pandoc`, je m'aperçois que le processus
    portant le PID 1 dans le container est `/bin/sh` (ou le shell de base).

Comparé le résultat de la commande `uname -rv` dans le container et sur la
machine hôte.

- Que remarquez-vous et que pouvez-vous en conclure ?
    - Dans le cas deux cas, le résultat de cette commande est `5.15.0-105-generic #115-Ubuntu SMP Mon Apr 15 09:52:04 UTC 2024`.
    On peut en conclure que le container est "conscient" de l'OS parent et que
    celui-ci sera le même pour lui (_maybe ?_).

Similairement à ci-dessus, comparez le contenu du répertoire `/dev/`.

- Que remarquez-vous ? À votre avis, quelle est la raison de cette différence ?
    - Dans le cas du container la liste est très petite comparé à la machine hôte.
    Une des choses manquantes en particulier sont les partitions de disque (p.ex.:
    `/dev/sda`, `/dev/sdb`). La raison de cette différence est que le container
    n'est "qu'un processus" par rapport à l'OS complète s'exécutant sur la
    machine hôte.

- Est-il possible de donner à un container accès à un ou plusieurs périphériques
de la machine hôte ? Comment ? Trouvez un moyen permettant de vérifier votre méthode.
    - Oui, il est possible de donner accès au container à un périphérique de
    la machine hôte à l'aide du paramètre `device`. Ci-dessous vous pouvez voir
    un exemple où on met à disposition du container, le device `/dev/sda` de
    la machine hôte. On retrouve par la suite ce device, lorsqu'on inspecte
    le retour de `ls -ltr /dev` dans le container.
    ```bash
    docker run --device=/dev/sda:/dev/host-sda -it --rm alpine:3.17
    ```

    ```bash
    brw-rw----    1 root     disk        8,   0 Apr 20 16:10 host-sda
    ```

Exécutez la commande `mount` sur la machine hôte et dans le container.

- Quelles différences remarquez-vous, notamment en ce qui concerne le système
de fichiers racine (`rootfs`) ?
